@@ -202,7 +202,7 @@ end
 
 -- Built-in defaults
 local core_defaults = {
-	capath = "/etc/ssl/certs";
+	cafile = "/etc/ssl/cert.pem";
 	depth = 9;
 	protocol = "tlsv1+";
 	verify = "none";
@@ -336,8 +336,11 @@ local function create_context(host, mode, ...)
 		password = function() log("error", "Encrypted certificate for %s requires 'ssl' 'password' to be set in config", host); end;
 	});
 	local profile = configmanager.get("*", "tls_profile") or "intermediate";
-	if profile ~= "legacy" then
+	if mozilla_ssl_configs[profile] then
 		cfg:apply(mozilla_ssl_configs[profile]);
+	elseif profile ~= "legacy" then
+		log("error", "Invalid value for 'tls_profile': expected one of \"modern\", \"intermediate\" (default), \"old\" or \"legacy\" but got %q", profile);
+		return nil, "Invalid configuration, 'tls_profile' had an unknown value.";
 	end
 	cfg:apply(global_ssl_config);
 
