@@ -87,7 +87,11 @@ local function on_retrieve_legacy_pep(event)
 	local service = mod_pep.get_pep_service(username);
 	local ok, ret = service:get_items(namespace, session.full_jid);
 	if not ok then
-		module:log("error", "Failed to retrieve PEP bookmarks of %s: %s", jid, ret);
+		if ret == "item-not-found" then
+			module:log("debug", "Got no PEP bookmarks item for %s, returning empty private bookmarks", jid);
+		else
+			module:log("error", "Failed to retrieve PEP bookmarks of %s: %s", jid, ret);
+		end
 		session.send(st.error_reply(stanza, "cancel", ret, "Failed to retrieve bookmarks from PEP"));
 		return true;
 	end
@@ -163,10 +167,15 @@ local function publish_to_pep(jid, bookmarks, synchronise)
 		if synchronise then
 			-- If we set zero legacy bookmarks, purge the bookmarks 2 node.
 			module:log("debug", "No bookmark in the set, purging instead.");
-			return service:purge(namespace, jid, true);
-		else
-			return true;
+			local ok, err = service:purge(namespace, jid, true);
+			-- It's okay if no node exists when purging, user has
+			-- no bookmarks anyway.
+			if not ok and err ~= "item-not-found" then
+				module:log("error", "Failed to clear items from bookmarks 2 node: %s", err);
+				return ok, err;
+			end
 		end
+		return true;
 	end
 
 	-- Retrieve the current bookmarks2.
@@ -305,7 +314,7 @@ local function on_publish_legacy_pep(event)
 
 	local ok, err = publish_to_pep(session.full_jid, bookmarks, true);
 	if not ok then
-		module:log("error", "Failed to publish to PEP bookmarks for %s@%s: %s", session.username, session.host, err);
+		module:log("error", "Failed to sync legacy bookmarks to PEP for %s@%s: %s", session.username, session.host, err);
 		session.send(st.error_reply(stanza, "cancel", "internal-server-error", "Failed to store bookmarks to PEP"));
 		return true;
 	end
@@ -331,7 +340,7 @@ local function on_publish_private_xml(event)
 
 	local ok, err = publish_to_pep(session.full_jid, bookmarks, true);
 	if not ok then
-		module:log("error", "Failed to publish to PEP bookmarks for %s@%s: %s", session.username, session.host, err);
+		module:log("error", "Failed to sync private XML bookmarks to PEP for %s@%s: %s", session.username, session.host, err);
 		session.send(st.error_reply(stanza, "cancel", "internal-server-error", "Failed to store bookmarks to PEP"));
 		return true;
 	end
