@@ -46,12 +46,19 @@ function startup.parse_args(profile)
 		end
 		os.exit(1);
 	end
-	if opts.help and prosody.process_type == "prosody" then
-		print("prosody [ -D | -F ] [ --config /path/to/prosody.cfg.lua ]");
-		print("  -D, --daemonize       Run in the background")
-		print("  -F, --no-daemonize    Run in the foreground")
-		print("  --config FILE         Specify config file")
-		os.exit(0);
+	if prosody.process_type == "prosody" then
+		if #arg > 0 then
+			print("Unrecognized option: "..arg[1]);
+			print("(Did you mean 'prosodyctl "..arg[1].."'?)");
+			print("");
+		end
+		if opts.help or #arg > 0 then
+			print("prosody [ -D | -F ] [ --config /path/to/prosody.cfg.lua ]");
+			print("  -D, --daemonize       Run in the background")
+			print("  -F, --no-daemonize    Run in the foreground")
+			print("  --config FILE         Specify config file")
+			os.exit(0);
+		end
 	end
 	prosody.opts = opts;
 end
@@ -125,6 +132,7 @@ end
 function startup.load_libraries()
 	-- Load socket framework
 	-- luacheck: ignore 111/server 111/socket
+	require "util.import"
 	socket = require "socket";
 	server = require "net.server"
 end
@@ -346,13 +354,12 @@ function startup.add_global_prosody_functions()
 			reason = reason;
 			code = code;
 		});
-		server.setquitting(true);
+		prosody.main_thread:run(startup.shutdown);
 	end
 end
 
 function startup.load_secondary_libraries()
 	--- Load and initialise core modules
-	require "util.import"
 	require "util.xmppstream"
 	require "core.stanza_router"
 	require "core.statsmanager"
@@ -503,7 +510,7 @@ function startup.switch_user()
 		local arg_root = prosody.opts.root;
 		if prosody.current_uid == 0 and config.get("*", "run_as_root") ~= true and not arg_root then
 			-- We haz root!
-			local desired_user = config.get("*", "prosody_user") or "prosody";
+			local desired_user = config.get("*", "prosody_user") or "_prosody";
 			local desired_group = config.get("*", "prosody_group") or desired_user;
 			local ok, err = pposix.setgid(desired_group);
 			if ok then
@@ -637,7 +644,11 @@ function startup.shutdown()
 
 	prosody.log("debug", "Shutdown reason was: %s", prosody.shutdown_reason or "not specified");
 	prosody.log("debug", "Exiting with status code: %d", prosody.shutdown_code or 0);
-	os.exit(prosody.shutdown_code);
+	server.setquitting(true);
+end
+
+function startup.exit()
+	os.exit(prosody.shutdown_code, true);
 end
 
 -- prosodyctl only
